import os
import scipy.io
import numpy as np
from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim





import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.nn.utils import spectral_norm
from torch.distributions.multivariate_normal import MultivariateNormal as multi_normal
from torchvision import utils as vutils
from itertools import chain
import numpy as np
from numpy import matrix, linalg, pi
import math
from scipy.stats import multivariate_normal

import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.nn.utils import spectral_norm




def append_params(params, module, prefix):
    for child in module.children():
        for k,p in child._parameters.items():
            if p is None: continue

            if isinstance(child, nn.BatchNorm2d):
                name = prefix + '_bn_' + k
            else:
                name = prefix + '_' + k

            if name not in params:
                params[name] = p
            else:
                raise RuntimeError('Duplicated param name: {:s}'.format(name))


def set_optimizer(model, lr_base, lr_mult, train_all=False, momentum=0.9, w_decay=0.0005):
    if train_all:
        params = model.get_all_params()
    else:
        params = model.get_learnable_params()
    param_list = []
    for k, p in params.items():
        lr = lr_base
        for l, m in lr_mult.items():
            if k.startswith(l):
                lr = lr_base * m
        param_list.append({'params': [p], 'lr':lr})
    optimizer = optim.SGD(param_list, lr = lr, momentum=momentum, weight_decay=w_decay)
    return optimizer

def set_requires_grad_3(m, requires_grad):
      pp_1=0
      lr_base=0.005
      momentum=0.9
      w_decay=0.0005
      param_list=[]
      for param in m.parameters():
         param.requires_grad_(requires_grad)
         param_list.append({'params':[param]})
         pp_1=pp_1+1
      optimizer = optim.SGD(param_list, lr = lr_base, momentum=momentum, weight_decay=w_decay)
      return optimizer



class ChannelAttentionMask(nn.Module):
    def __init__(self, c_dim):
        super().__init__()
#        self.feat_c = feat_c
#        self.feat_hw = feat_hw

        self.instance_attention = nn.Parameter(torch.randn(1,3,107))
        self.channel_attention = nn.Sequential(
            nn.Linear(c_dim, 3), nn.ReLU(), nn.Linear(3, 3), UnFlatten(1)
        )
    def forward(self, c):
        #instance_mask = torch.softmax(self.instance_attention, dim=-1).view(1, self.feat_c, self.feat_hw, self.feat_hw)
        channel_mask = self.channel_attention(c)
        return 3*channel_mask
    
    
class UnFlatten(nn.Module):
    def __init__(self, block_size):
        super(UnFlatten, self).__init__()
        self.block_size = block_size

    def forward(self, x):
        return x.view(x.size(0), -1, self.block_size, self.block_size)




class OOGANInput(nn.Module):
    """
    The OOGAN input module, the compteting-free input
    ...

    Attributes
    ----------
    c_dim : int
        number of dimensions in control vector c
    z_dim : int
        number of dimensions in noise vector z
    feat_4 : int
        feature's channel dimension at 4x4 level
    feat_8 : int
        feature's channel dimension at 8x8 level

    Methods
    -------
    forward(c=None, z=None): Tensor
        returns the feature map
    """
    def __init__(self, cc,patch):
        super().__init__()

        
        self.from_c_4 = nn.Sequential(
            UnFlatten(1), 
            nn.ConvTranspose2d(cc, 3, 107, 1, 0, bias=True),
            nn.BatchNorm2d(3))
#        nn.LeakyReLU(0.01)
        self.patch=patch
        self.attn_from_c = ChannelAttentionMask(cc)


    def forward(self,cc,patch):
       
        feat = self.from_c_4(cc)
#        attn_from_c = self.attn_from_c(cc)
        feat = patch + feat
        return feat
    



class MDNet(nn.Module):
    def __init__(self, model_path=None, K=1):
        super(MDNet, self).__init__()
        self.K = K
        self.layers = nn.Sequential(OrderedDict([
                ('conv1', nn.Sequential(nn.Conv2d(3, 96, kernel_size=7, stride=2),
                                        nn.ReLU(inplace=True),
                                        nn.LocalResponseNorm(2),
                                        nn.MaxPool2d(kernel_size=3, stride=2))),
                ('conv2', nn.Sequential(nn.Conv2d(96, 256, kernel_size=5, stride=2),
                                        nn.ReLU(inplace=True),
                                        nn.LocalResponseNorm(2),
                                        nn.MaxPool2d(kernel_size=3, stride=2))),
                ('conv3', nn.Sequential(nn.Conv2d(256, 512, kernel_size=3, stride=1),
                                        nn.ReLU(inplace=True))),
                ('fc4',   nn.Sequential(nn.Linear(512 * 3 * 3, 512),
                                        nn.ReLU(inplace=True))),
                ('fc5',   nn.Sequential(nn.Dropout(0.5),
                                        nn.Linear(512, 512),
                                        nn.ReLU(inplace=True)))]))

        self.branches = nn.ModuleList([nn.Sequential(nn.Dropout(0.5),
                                                     nn.Linear(512, 2)) for _ in range(K)])
        
    
    
        for m in self.layers.modules():
            if isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0.1)
        for m in self.branches.modules():
            if isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)
         
        
        if model_path is not None:
            if os.path.splitext(model_path)[1] == '.pth':
                self.load_model(model_path)
            elif os.path.splitext(model_path)[1] == '.mat':
                self.load_mat_model(model_path)
            else:
                raise RuntimeError('Unkown model format: {:s}'.format(model_path))
        self.build_param_dict()

    def build_param_dict(self):
        self.params = OrderedDict()
        for name, module in self.layers.named_children():
            append_params(self.params, module, name)
        for k, module in enumerate(self.branches):
            append_params(self.params, module, 'fc6_{:d}'.format(k))
        

    def set_learnable_params(self, layers):
        for k, p in self.params.items():
            if any([k.startswith(l) for l in layers]):
                p.requires_grad = True
            else:
                p.requires_grad = False
                

    def get_learnable_params(self):
        params = OrderedDict()
        for k, p in self.params.items():
            if p.requires_grad:
                params[k] = p
        return params
    
    def get_all_params(self):
        params = OrderedDict()
        for k, p in self.params.items():
            params[k] = p
        return params

    def forward(self, x, k=0,id_x=None, in_layer='conv1', out_layer=''):
#        self.K = K
        
        # forward model from in_layer to out_layer
        
        
        run = False
        for name, module in self.layers.named_children():
            if name == in_layer:
                run = True
            if run:
                x = module(x)
                if name == 'conv3':
                    x = x.view(x.size(0), -1)
                if name == out_layer:
                    return x

        x = self.branches[k](x)
        if out_layer=='fc6':
            return x
        elif out_layer=='fc6_softmax':
            return F.softmax(x, dim=1)
        
        
        
        
        
    

    def load_model(self, model_path):
        states = torch.load(model_path)
#        shared_layers = states['shared_layers']       
        self.layers.load_state_dict(states)
        

    def load_mat_model(self, matfile):
        mat = scipy.io.loadmat(matfile)
        mat_layers = list(mat['layers'])[0]

        # copy conv weights
        for i in range(3):
            weight, bias = mat_layers[i * 4]['weights'].item()[0]
            self.layers[i][0].weight.data = torch.from_numpy(np.transpose(weight, (3, 2, 0, 1)))
            self.layers[i][0].bias.data = torch.from_numpy(bias[:, 0])

class QHead(nn.Module):
    def __init__(self,model_path=None):
        super().__init__()

#        self.conv1 = nn.Conv2d(4608, 128,1,bias=False)
#        self.bn1 = nn.BatchNorm2d(32)
#        self.bn2 = nn.BatchNorm2d(96)
#        self.conv2=nn.Conv2d(128,2,1)
#        self.conv3=nn.Conv2d(128,2,1)
#        self.act=nn.ReLU(inplace=True)
#        self.bn1 = nn.BatchNorm1d(128)

        
#        self.conv_fore = nn.Linear(128, 2)
#        self.conv_back = nn.Linear(128, 2)



        self.conv1=nn.Linear(512,128)
        self.conv2=nn.Linear(128,2)
        self.conv3=nn.Linear(128,2)
        
        if model_path is not None:
            if os.path.splitext(model_path)[1] == '.pth':
                self.load_model(model_path)  


    def forward(self, x,id_x=None):
#        yy=UnFlatten(3)
#        yy_1=yy(x)
        if id_x==0:
            x = F.leaky_relu(self.conv2((self.conv1(x))), 0.1, inplace=True)
#            xx = self.act(self.conv2(self.conv1(x)))
        elif id_x==1:
#            xx = self.act(self.conv3(self.conv1(x)))
            x = F.leaky_relu(self.conv3((self.conv1(x))), 0.1, inplace=True)
            
#        return xx[:,0],xx[:,1]
#        mu=torch.zeros((xx.size(0),c_dim))
#        var=torch.zeros((xx.size(0),c_dim))
#        
#           
#        if id_x==0:
#            x_1 = self.conv_fore(xx)
#            xx_1=torch.clamp(x_1, min=0, max=1)
#            x_2=xx_1.squeeze()
#            
#            x_3 = torch.exp(xx_1[:,1])
#            x_4=x_3.squeeze()
#            
#        
#        if id_x==1:
#            x_1 = self.conv_back(xx)
#            xx_1=torch.clamp(x_1, min=0, max=1)
#            x_2=xx_1.squeeze()
#            
#            x_3 = torch.exp(xx_1[:,1])
#            x_4=x_3.squeeze()
#            
#        for ii in range(xx.size(0)):
#            for jj in range(c_dim):        
#                mu[ii,jj]=xx[ii,0]
#                var[ii,jj]=xx[ii,1]
##                
#        mu=mu.cuda()  
#        var=var.cuda() 
#                
        return  x[:,0],x[:,1]
    
    def load_model(self, model_path):
        states_2 = torch.load(model_path)
#        shared_layers = states_2['shared_layers_2']       
        self.load_state_dict(states_2)
        
       




class BCELoss(nn.Module):
    def forward(self, pos_score, neg_score, average=True):
        pos_loss = -F.log_softmax(pos_score, dim=1)[:, 1]
        pos_loss_p = (torch.ones(pos_loss.size()).cuda() - F.softmax(pos_score, dim=1)[:,1]) * pos_loss
        neg_loss = -F.log_softmax(neg_score, dim=1)[:, 0]
        neg_loss_p = (torch.ones(neg_loss.size()).cuda() - F.softmax(neg_score, dim=1)[:,0]) * neg_loss

        loss = pos_loss_p.sum() + neg_loss_p.sum()
        # loss = pos_loss.sum() + neg_loss.sum()
        if average:
            # loss /= (pos_loss.size(0) + neg_loss.size(0))
            loss /= (pos_loss_p.size(0) + neg_loss_p.size(0))
        return loss

 
class Neg_log_density():
        def __call__(self, x,mu,var):
            constant = torch.Tensor([np.log(2 * np.pi)])
            constant=constant.cuda()
            mu = mu
            logsigma = var
            inv_sigma = torch.exp(-logsigma)
            tmp = (x - mu) * inv_sigma
            return  0.5 * (tmp * tmp + 2 * logsigma + constant)
            
#            logli = -0.5 * (var.mul(2 * np.pi) + 1e-6).log() - (x - mu).pow(2).div(var.mul(2.0) + 1e-6)
#            nll = -(logli.sum(1).mean())

#            return nll


class Accuracy():
    def __call__(self, pos_score, neg_score):
        pos_correct = (pos_score[:, 1] > pos_score[:, 0]).sum().float()
        neg_correct = (neg_score[:, 1] < neg_score[:, 0]).sum().float()
        acc = (pos_correct + neg_correct) / (pos_score.size(0) + neg_score.size(0) + 1e-8)
        return acc.item()


class Precision():
    def __call__(self, pos_score, neg_score):
        scores = torch.cat((pos_score[:, 1], neg_score[:, 1]), 0)
        topk = torch.topk(scores, pos_score.size(0))[1]
        prec = (topk < pos_score.size(0)).float().sum() / (pos_score.size(0) + 1e-8)
        return prec.item()
